@@ -25,18 +25,27 @@ import json
 
 from pathlib import Path
 from threading import Thread
+from typing import Union
 from socketio import ClientNamespace, Client as SocketIO
 from gevent.pywsgi import WSGIServer
 from enum import Enum
 
-from vantage6.common.docker_addons import ContainerKillListener
+from vantage6.common.docker.addons import (
+    ContainerKillListener, check_docker_running, running_in_docker
+)
 from vantage6.common.globals import VPN_CONFIG_FILE
-from vantage6.node.globals import NODE_PROXY_SERVER_HOSTNAME
+from vantage6.common.exceptions import AuthenticationException
+from vantage6.cli.context import NodeContext
+from vantage6.node.context import DockerNodeContext
+from vantage6.node.globals import (
+    NODE_PROXY_SERVER_HOSTNAME, SLEEP_BTWN_NODE_LOGIN_TRIES,
+    TIME_LIMIT_RETRY_CONNECT_NODE, TIME_LIMIT_INITIAL_CONNECTION_WEBSOCKET
+)
 from vantage6.node.server_io import NodeClient
-from vantage6.node.proxy_server import app
+from vantage6.node import proxy_server
 from vantage6.node.util import logger_name
 from vantage6.node.docker.docker_manager import DockerManager
-from vantage6.node.docker.network_manager import IsolatedNetworkManager
+from vantage6.common.docker.network_manager import NetworkManager
 from vantage6.node.docker.vpn_manager import VPNManager
 
 
@@ -62,6 +71,12 @@ class NodeTaskNamespace(ClientNamespace):
     def on_message(self, data):
         self.log.info(data)
 
+    def on_connect(self):
+        """On connect or reconnect"""
+        self.log.info('(Re)Connected to the /tasks namespace')
+        self.node_worker_ref.sync_task_queue_with_server()
+        self.log.debug("Tasks synced again with the server...")
+
     def on_disconnect(self):
         """ Server disconnects event."""
         # self.node_worker_ref.socketIO.disconnect()
@@ -97,10 +112,8 @@ class NodeTaskNamespace(ClientNamespace):
         self.log.debug("Token refreshed")
         self.node_worker_ref.connect_to_socket()
         self.log.debug("Connected to socket")
-
-        # FIXME: This won't work: you're trying to access a private method!?
-        # self.node_worker_ref.__sync_task_queue_with_server()
-        # self.log.debug("Tasks synced again with the server...")
+        self.node_worker_ref.sync_task_queue_with_server()
+        self.log.debug("Tasks synced again with the server...")
 
 
 # ------------------------------------------------------------------------------
@@ -133,11 +146,14 @@ class Node(object):
         # already
         try:
             self.initialize()
-        except Exception as e:
+        except Exception:
             self.cleanup()
             raise
 
     def initialize(self):
+        # check if docker is running, otherwise exit with error
+        check_docker_running()
+
         self.config = self.ctx.config
         self.queue = queue.Queue()
         self._using_encryption = None
@@ -169,13 +185,15 @@ class Node(object):
         self.log.debug("Creating websocket connection with the server")
         self.connect_to_socket()
 
-        # Check if new tasks were posted while offline.
-        self.log.debug("Fetching tasks that were posted while offline")
-        self.__sync_task_queue_with_server()
-
         # setup docker isolated network manager
-        isolated_network_mgr = \
-            IsolatedNetworkManager(self.ctx.docker_network_name)
+        internal_ = running_in_docker()
+        if not internal_:
+            self.log.warn(
+                "Algorithms have internet connection! "
+                "This happens because you use 'vnode-local'!"
+            )
+        isolated_network_mgr = NetworkManager(self.ctx.docker_network_name)
+        isolated_network_mgr.create_network(is_internal=internal_)
 
         # Setup tasks dir
         self._set_task_dir(self.ctx)
@@ -201,6 +219,10 @@ class Node(object):
                 aliases=[NODE_PROXY_SERVER_HOSTNAME]
             )
 
+        # Connect any docker services specified in the configuration file to
+        # the node container
+        self.link_docker_services()
+
         # Thread for sending results to the server when they come available.
         self.log.debug("Start thread for sending messages (results)")
         t = Thread(target=self.__speaking_worker, daemon=True)
@@ -220,12 +242,6 @@ class Node(object):
             A proxy for communication between algorithms and central
             server.
         """
-        # supply the proxy server with a destination (the central server)
-        # we might want to not use enviroment vars
-        os.environ["SERVER_URL"] = self.server_io.host
-        os.environ["SERVER_PORT"] = self.server_io.port
-        os.environ["SERVER_PATH"] = self.server_io.path
-
         if self.ctx.running_in_docker:
             # NODE_PROXY_SERVER_HOSTNAME points to the name of the proxy
             # when running in the isolated docker network.
@@ -245,13 +261,14 @@ class Node(object):
 
         # 'app' is defined in vantage6.node.proxy_server
         # app.debug = True
-        app.config["SERVER_IO"] = self.server_io
+        proxy_server.app.config["SERVER_IO"] = self.server_io
+        proxy_server.server_url = self.server_io.base_path
 
         # this is where we try to find a port for the proxyserver
         for try_number in range(5):
             self.log.info(
                 f"Starting proxyserver at '{proxy_host}:{proxy_port}'")
-            http_server = WSGIServer(('0.0.0.0', proxy_port), app)
+            http_server = WSGIServer(('0.0.0.0', proxy_port), proxy_server.app)
 
             try:
                 http_server.serve_forever()
@@ -273,7 +290,7 @@ class Node(object):
                 self.log.error('Proxyserver could not be started or crashed!')
                 self.log.error(e)
 
-    def __sync_task_queue_with_server(self):
+    def sync_task_queue_with_server(self):
         """ Get all unprocessed tasks from the server for this node."""
         assert self.server_io.cryptor, "Encrpytion has not been setup"
 
@@ -319,21 +336,26 @@ class Node(object):
 
         # Run the container. This adds the created container/task to the list
         # __docker.active_tasks
-        vpn_port = self.__docker.run(
+        vpn_ports = self.__docker.run(
             result_id=taskresult["id"],
             image=task["image"],
             docker_input=taskresult['input'],
             tmp_vol_name=vol_name,
-            token=token
+            token=token,
+            database=task.get('database', 'default')
         )
 
-        if vpn_port:
+        if vpn_ports:
             # Save port of VPN client container at which it redirects traffic
-            # to the algorithm container
+            # to the algorithm container. First delete any existing port
+            # assignments in case algorithm has crashed
             self.server_io.request(
-                f"result/{taskresult['id']}", json={"port": vpn_port},
-                method="PATCH"
+                'port', params={'result_id': taskresult['id']}, method="DELETE"
             )
+            for port in vpn_ports:
+                port['result_id'] = taskresult['id']
+                self.server_io.request('port', method='POST', json=port)
+
             # Save IP address of VPN container
             node_id = self.server_io.whoami.id_
             node_ip = self.vpn_manager.get_vpn_ip()
@@ -347,7 +369,7 @@ class Node(object):
             Runs in a separate thread. Received events are dispatched
             through the appropriate action_handler for a channel.
         """
-        self.log.debug("listening for incoming messages")
+        self.log.debug("Listening for incoming messages")
 
         # FIXME: while True in combination with a wait() call that never exits
         #   makes joining the tread (to terminate) difficult?
@@ -355,7 +377,11 @@ class Node(object):
             # incoming messages are handled by the action_handler instance
             # which is attached when the socket connection was made. wait()
             # is blocks forever (if no time is specified).
-            self.socketIO.wait()
+            try:
+                self.socketIO.wait()
+            except Exception as e:
+                self.log.error('Listening thread had an exception')
+                self.log.debug(e)
 
     def __speaking_worker(self):
         """ Sending messages to central server.
@@ -369,54 +395,61 @@ class Node(object):
         self.log.debug("Waiting for results to send to the server")
 
         while True:
-            results = self.__docker.get_result()
+            try:
+                results = self.__docker.get_result()
 
-            # notify all of a crashed container
-            if results.status_code:
-                self.socketIO.emit(
-                    'container_failed',
-                    data={
-                        'node_id': self.server_io.whoami.id_,
-                        'status_code': results.status_code,
-                        'result_id': results.result_id,
-                        'collaboration_id': self.server_io.collaboration_id
-                    },
-                    namespace='/tasks'
+                # notify all of a crashed container
+                if results.status_code:
+                    self.socketIO.emit(
+                        'container_failed',
+                        data={
+                            'node_id': self.server_io.whoami.id_,
+                            'status_code': results.status_code,
+                            'result_id': results.result_id,
+                            'collaboration_id': self.server_io.collaboration_id
+                        },
+                        namespace='/tasks'
+                    )
+
+                self.log.info(
+                    f"Sending result (id={results.result_id}) to the server!")
+
+                # FIXME: why are we retrieving the result *again*? Shouldn't we
+                # just store the task_id when retrieving the task the first
+                # time?
+                response = self.server_io.request(
+                    f"result/{results.result_id}"
                 )
+                task_id = response.get("task").get("id")
 
-            self.log.info(
-                f"Sending result (id={results.result_id}) to the server!")
+                if not task_id:
+                    self.log.error(
+                        f"task_id of result (id={results.result_id}) "
+                        f"could not be retrieved"
+                    )
+                    return
 
-            # FIXME: why are we retrieving the result *again*? Shouldn't we
-            # just store the task_id when retrieving the task the first time?
-            response = self.server_io.request(f"result/{results.result_id}")
-            task_id = response.get("task").get("id")
+                response = self.server_io.request(f"task/{task_id}")
+                initiator_id = response.get("initiator")
 
-            if not task_id:
-                self.log.error(
-                    f"task_id of result (id={results.result_id}) "
-                    f"could not be retrieved"
+                if not initiator_id:
+                    self.log.error(
+                        f"Initiator id from task (id={task_id})could not be "
+                        f"retrieved"
+                    )
+
+                self.server_io.patch_results(
+                    id=results.result_id,
+                    initiator_id=initiator_id,
+                    result={
+                        'result': results.data,
+                        'log': results.logs,
+                        'finished_at': datetime.datetime.now().isoformat(),
+                    }
                 )
-                return
-
-            response = self.server_io.request(f"task/{task_id}")
-            initiator_id = response.get("initiator")
-
-            if not initiator_id:
-                self.log.error(
-                    f"Initiator id from task (id={task_id})could not be "
-                    f"retrieved"
-                )
-
-            self.server_io.patch_results(
-                id=results.result_id,
-                initiator_id=initiator_id,
-                result={
-                    'result': results.data,
-                    'log': results.logs,
-                    'finished_at': datetime.datetime.now().isoformat(),
-                }
-            )
+            except Exception as e:
+                self.log.error('Speaking thread had an exception')
+                self.log.debug(e)
 
     def authenticate(self):
         """ Authenticate to the central server
@@ -426,23 +459,34 @@ class Node(object):
         """
         api_key = self.config.get("api_key")
 
-        keep_trying = True
-        while keep_trying:
+        success = False
+        i = 0
+        while i < TIME_LIMIT_RETRY_CONNECT_NODE / SLEEP_BTWN_NODE_LOGIN_TRIES:
+            i = i + 1
             try:
                 self.server_io.authenticate(api_key)
 
+            except AuthenticationException as e:
+                msg = "Authentication failed: API key is wrong!"
+                self.log.warning(msg)
+                self.log.debug(e)
+                break
             except Exception as e:
                 msg = 'Authentication failed. Retrying in 10 seconds!'
                 self.log.warning(msg)
                 self.log.debug(e)
-                time.sleep(10)
+                time.sleep(SLEEP_BTWN_NODE_LOGIN_TRIES)
 
             else:
                 # This is only executed if try-block executed without error.
-                keep_trying = False
+                success = True
+                break
 
-        # At this point, we shoud be connnected.
-        self.log.info(f"Node name: {self.server_io.name}")
+        if success:
+            self.log.info(f"Node name: {self.server_io.name}")
+        else:
+            self.log.critical('Unable to authenticate. Exiting')
+            exit(1)
 
     def private_key_filename(self):
         """Get the path to the private key."""
@@ -523,11 +567,20 @@ class Node(object):
             self.__vpn_dir = ctx.vpn_dir
 
     def setup_vpn_connection(
-            self, isolated_network_mgr: IsolatedNetworkManager, ctx
+        self,
+        isolated_network_mgr: NetworkManager,
+        ctx: Union[DockerNodeContext, NodeContext]
     ) -> VPNManager:
         """
         Setup container which has a VPN connection
 
+        Parameters
+        ----------
+        isolated_network_mgr: NetworkManager
+            Manager for the isolated network
+        ctx: NodeContext
+            Context object for the node
+
         Returns
         -------
         VPNManager
@@ -535,17 +588,37 @@ class Node(object):
         """
         ovpn_file = os.path.join(self.__vpn_dir, VPN_CONFIG_FILE)
 
-        self.log.debug("Setting up VPN client container")
+        self.log.info("Setting up VPN client container")
         vpn_volume_name = self.ctx.docker_vpn_volume_name \
             if ctx.running_in_docker else self.__vpn_dir
+
+        #FIXME: remove me in 4+. alpine image has been moved into the `images`
+        # key. This is to support older configuration files.
+        legacy_alpine = self.config.get('alpine')
+
+        # user can specify custom images in the configuration file
+        custom_alpine = self.config['images'].get('alpine') \
+            if 'images' in self.config else None
+        custom_vpn_client = self.config['images'].get('vpn_client') \
+            if 'images' in self.config else None
+        custom_network = self.config['images'].get('network_config') \
+            if 'images' in self.config else None
+
         vpn_manager = VPNManager(
             isolated_network_mgr=isolated_network_mgr,
             node_name=self.ctx.name,
             vpn_volume_name=vpn_volume_name,
-            vpn_subnet=self.config.get('vpn_subnet')
+            vpn_subnet=self.config.get('vpn_subnet'),
+            alpine_image=custom_alpine or legacy_alpine,
+            vpn_client_image=custom_vpn_client,
+            network_config_image=custom_network
         )
-        # if vpn config doesn't exist, get it and write to disk
-        if not os.path.isfile(ovpn_file):
+
+        if not self.config.get('vpn_subnet'):
+            self.log.warn("VPN subnet is not defined!")
+            self.log.info("Not trying to establish VPN connection.")
+        elif not os.path.isfile(ovpn_file):
+            # if vpn config doesn't exist, get it and write to disk
             self._connect_vpn(vpn_manager, VPNConnectMode.REFRESH_COMPLETE,
                               ovpn_file)
         else:
@@ -622,6 +695,16 @@ class Node(object):
             f.write(ovpn_config)
         return True
 
+    def link_docker_services(self):
+        docker_services = self.ctx.config.get("docker_services")
+        if not docker_services:
+            return
+        self.log.info("Linking docker services specified in the configuration")
+        for alias, container_name in docker_services.items():
+            self.__docker.link_container_to_network(
+                container_name=container_name, config_alias=alias
+            )
+
     def connect_to_socket(self):
         """ Create long-lasting websocket connection with the server.
 
@@ -630,26 +713,29 @@ class Node(object):
         """
 
         self.socketIO = SocketIO(request_timeout=60)
-        self.socketIO.connect(
-            url=f'{self.server_io.host}:{self.server_io.port}',
-            headers=self.server_io.headers,
-            namespaces=['/tasks']
-        )
 
         self.socketIO.register_namespace(NodeTaskNamespace('/tasks'))
         NodeTaskNamespace.node_worker_ref = self
 
-        # Log the outcome
-        while not self.socketIO.connected:
-            self.log.debug('waiting for socket connection...')
-            time.sleep(1)
-
-        msg = 'connected to host={host} on port={port}'
-        msg = msg.format(
-            host=self.server_io.host,
-            port=self.server_io.port
+        self.socketIO.connect(
+            url=f'{self.server_io.host}:{self.server_io.port}',
+            headers=self.server_io.headers,
+            wait=False
         )
-        self.log.info(msg)
+
+        # Log the outcome
+        i = 0
+        while not self.socketIO.connected:
+            if i > TIME_LIMIT_INITIAL_CONNECTION_WEBSOCKET:
+                self.log.critical('Could not connect to the websocket '
+                                  'channels, do you have a slow connection?')
+                exit(1)
+            self.log.debug('Waiting for socket connection...')
+            time.sleep(1)
+            i += 1
+
+        self.log.info(f'Connected to host={self.server_io.host} on port='
+                      f'{self.server_io.port}')
 
     def get_task_and_add_to_queue(self, task_id):
         """Fetches (open) task with task_id from the server.
